NullIntersectors := function(G, A)

local   sgl, # list of all subgroups of G
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
      maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice
	     # (see MinimalSupergroupsLattice and MaximalSubgroupsLattice in the Reference Manual for details
     marked, # list used for keeping track of already "visited" subgroups (vertices) and the direction of 
             # the Breadth-First Search propagation
             # marked[x] < 0 - the node x is already reached ("visited")
	     # marked[x] > 0 - the node x has not been visited yet
	     # |marked[x]| = 2 - x is a subgroup of A (so we can keep going to upper levels)
	     # |marked[x]| = 1 - x is not a subgroup of A
      queue, # list used to simulate a queue data structure
          N, # the function N(x) returns the the neighbourhood of x  
a, x, sg, t, result, S, c; # auxiliary variables used in for loops, etc.

# this function returns the neighbourhood of the subgroup x
# it is assumed that the minimal supergroups and maximal subgroups relations are built in minsl resp. maxsl
# if x is an subgroup of A, then this function returns both the minimal supergroups and the maximal subgroups of x
# otherwise it returns only a list with the minimal supergroups
# it also marks the subgroups of A and isolates the vertex corresponding to the trivial subgroup
N := function(x)
local minslx, maxslx, result;
  result := [];
  if AbsInt(marked[x]) = 2 then
    maxslx := maxsl[x]{[1..Size(maxsl[x])]}[1];
    marked{maxslx} := List([1..Size(maxslx)], c -> SignInt(c) * 2);
    if not maxslx = [ t ] then result := maxslx; fi;
  fi;
  Append(result, minsl[x]{[1..Size(minsl[x])]}[1]);
  return result;
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: NullIntersectors(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# if A is essential then 0 is the only solution
if IsEssential(G,A) then 
  return [ TrivialSubgroup(G) ];
fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# if A = 0 then for all subgroups H of G, H \cap 0 = 0.
if A = TrivialSubgroup(G) then
  return sgl;
fi;

# compute the minimal supergroups relation
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));
# compute the maximal subgroups relation
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

a := Position(sgl, A);
t := Position(sgl, TrivialSubgroup(G));

# Initialize the "marked" list
marked := List([1..Size(sgl)], c -> 1);

# A is a subgroup of itself
marked[a] := 2;

# A is also the first vertex being visited
queue := [ a ];
marked[a] := - marked[a];

# S will contain a list of all vertices in which we can get from A without passing through 0
# it is clear that for all H in S we have $ A \cap H \neq G $
S := [];

# Breadth-First Search for using the special neighbourhood function N(x)
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(S, x);
  for sg in Filtered(N(x), function(sg) return marked[sg] > 0; end) do
    Add(queue, sg);
    marked[sg] := - marked[sg];
  od;
od;

# the final result is obtained by subtracting from the set of all subgroups the set S
result := [1..Size(sgl)];
SubtractSet(result, S);

return sgl{result};

end;
