##########################################################################################################
## Returns a list of all coessential subgroups of A in G. 

IsClosed := function(G, A)

local minsl, sgl, t, a;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: IsClosed(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

if A = G then return true; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);


minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));
t := Position(sgl, TrivialSubgroup(G));
a := Position(sgl, A);

return Size(Filtered(minsl[t]{[1..Size(minsl[t])]}[1], c -> IsSubset(A, sgl[c]))) < 
       Minimum(List(minsl[a]{[1..Size(minsl[a])]}[1], c -> Size(Filtered(minsl[t]{[1..Size(minsl[t])]}[1], d -> IsSubset(sgl[c], sgl[d])))));

end;

##########################################################################################################
## Returns true if A is a direct summand of G, false otherwise.

IsDirectSummand := function(G, A)

local sgl, B, d;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: IsDirectSummand(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# if A = G then return false; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

d := false;

for B in Filtered(sgl, function(g) return Size(g) = IndexNC(G, A); end) do
  if Intersection(A, B) = TrivialSubgroup(G) then 
    d := true;
    break;
  fi;
od;

return d;

end;

##########################################################################################################
## Returns true if H is an essential subgroup of G, false otherwise.

IsEssential := function(G, H)

#ensure that G and H are finite abelian groups and H is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, H) then
  Error("Usage: IsEssential(<G>, <H>) - where <G> is a finite abelian group and <H> is a subgroup of <G>\n");
fi;

if G = H then return true;
	 else return IsSubset(H, Socle(G));
fi;

end;

##########################################################################################################
## Returns true if H is a superfluous subgroup of G, false otherwise.

IsSuperfluous := function(G, H)

#ensure that G and H are finite abelian groups and H is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, H) then
  Error("Usage: IsSuperfluous(<G>, <H>) - where <G> is a finite abelian group and <H> is a subgroup of <G>\n");
fi;

if G = TrivialSubgroup(G) then return true;  
			  else return IsSubset(Intersection(MaximalSubgroups(G)), H);
fi;

end;

##########################################################################################################
## Returns true if G is uniform, false otherwise. 

IsUniform := function(G)

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: IsUniform(<G>) - where <G> is a finite abelian group\n");
fi;

return Size(MaximalSubgroups(G)) = 1;

end;

##########################################################################################################
## Returns a list of subgroups of G such that their sum with A gives G.

FullSummands := function(G, A)

local   sgl, # list of all subgroups of G
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
      maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice
	     # (see MinimalSupergroupsLattice and MaximalSubgroupsLattice in the Reference Manual for details
     marked, # list used for keeping track of already "visited" subgroups (vertices) and the direction of 
             # the Breadth-First Search propagation
             # marked[x] < 0 - the node x is already reached ("visited")
	     # marked[x] > 0 - the node x has not been visited yet
	     # |marked[x]| = 2 - x is an extension of A (so we can keep to upper levels)
	     # |marked[x]| = 1 - x is not an extension of A
      queue, # list used to simulate a queue data structure
          N, # the function N(x) returns the the neighbourhood of x  
a, x, sg, g, result, S, c; # auxiliary variables used in for loops, etc.

# this function returns the neighbourhood of the subgroup x
# it is assumed that the minimal supergroups and maximal subgroups relations are built in minsl resp. maxsl
# if x is an extension of A, then this function returns both the minimal supergroups and the maximal subgroups of x
# otherwise it returns only a list with the maximal subgroups
# it also marks the supergroups of A and isolates the vertex corresponding to G
N := function(x)
local minslx, maxslx, result;
  result := [];
  if AbsInt(marked[x]) = 2 then
    minslx := minsl[x]{[1..Size(minsl[x])]}[1];
    marked{minslx} := List([1..Size(minslx)], c -> SignInt(c) * 2);
    if not minslx = [ g ] then result := minslx; fi;
  fi;
  Append(result, maxsl[x]{[1..Size(maxsl[x])]}[1]);
  return result;
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: FullSummands(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# if A is superfluous then then A + G = G is the only solution
if IsSuperfluous(G,A) then 
  return [ G ];
fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# if A = G then for all subgroups H of G, G = G + H.
if A = G then
  return sgl;
fi;

# compute the minimal supergroups relation
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));
# compute the maximal subgroups relation
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

a := Position(sgl, A);
g := Position(sgl, G);

# Initialize the "marked" list
marked := List([1..Size(sgl)], c -> 1);

# A is a supergroup of itself
marked[a] := 2;

# A is also the first vertex being visited
queue := [ a ];
marked[a] := - marked[a];

# S will contain a list of all vertices in which we can get from A without passing through G
# it is clear that for all H in S we have $ A + H \neq G $
S := [];

# Breadth-First Search for using the special neighbourhood function N(x)
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(S, x);
  for sg in Filtered(N(x), function(sg) return marked[sg] > 0; end) do
    Add(queue, sg);
    marked[sg] := - marked[sg];
  od;
od;

# the final result is obtained by subtracting from the set of all subgroups the set S
result := [1..Size(sgl)];
SubtractSet(result, S);

return sgl{result};

end;

##########################################################################################################
## Returns a list of subgroups of G such that their intersection with A is the trivial subgroup.

NullIntersectors := function(G, A)

local   sgl, # list of all subgroups of G
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
      maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice
	     # (see MinimalSupergroupsLattice and MaximalSubgroupsLattice in the Reference Manual for details
     marked, # list used for keeping track of already "visited" subgroups (vertices) and the direction of 
             # the Breadth-First Search propagation
             # marked[x] < 0 - the node x is already reached ("visited")
	     # marked[x] > 0 - the node x has not been visited yet
	     # |marked[x]| = 2 - x is a subgroup of A (so we can keep going to upper levels)
	     # |marked[x]| = 1 - x is not a subgroup of A
      queue, # list used to simulate a queue data structure
          N, # the function N(x) returns the the neighbourhood of x  
a, x, sg, t, result, S, c; # auxiliary variables used in for loops, etc.

# this function returns the neighbourhood of the subgroup x
# it is assumed that the minimal supergroups and maximal subgroups relations are built in minsl resp. maxsl
# if x is an subgroup of A, then this function returns both the minimal supergroups and the maximal subgroups of x
# otherwise it returns only a list with the minimal supergroups
# it also marks the subgroups of A and isolates the vertex corresponding to the trivial subgroup
N := function(x)
local minslx, maxslx, result;
  result := [];
  if AbsInt(marked[x]) = 2 then
    maxslx := maxsl[x]{[1..Size(maxsl[x])]}[1];
    marked{maxslx} := List([1..Size(maxslx)], c -> SignInt(c) * 2);
    if not maxslx = [ t ] then result := maxslx; fi;
  fi;
  Append(result, minsl[x]{[1..Size(minsl[x])]}[1]);
  return result;
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: NullIntersectors(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# if A is essential then 0 is the only solution
if IsEssential(G,A) then 
  return [ TrivialSubgroup(G) ];
fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# if A = 0 then for all subgroups H of G, H \cap 0 = 0.
if A = TrivialSubgroup(G) then
  return sgl;
fi;

# compute the minimal supergroups relation
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));
# compute the maximal subgroups relation
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

a := Position(sgl, A);
t := Position(sgl, TrivialSubgroup(G));

# Initialize the "marked" list
marked := List([1..Size(sgl)], c -> 1);

# A is a subgroup of itself
marked[a] := 2;

# A is also the first vertex being visited
queue := [ a ];
marked[a] := - marked[a];

# S will contain a list of all vertices in which we can get from A without passing through 0
# it is clear that for all H in S we have $ A \cap H \neq G $
S := [];

# Breadth-First Search for using the special neighbourhood function N(x)
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(S, x);
  for sg in Filtered(N(x), function(sg) return marked[sg] > 0; end) do
    Add(queue, sg);
    marked[sg] := - marked[sg];
  od;
od;

# the final result is obtained by subtracting from the set of all subgroups the set S
result := [1..Size(sgl)];
SubtractSet(result, S);

return sgl{result};

end;

##########################################################################################################
## Returns a list of all essential subgroups of G.

EssentialSubgroups := function(G)
local   sgl, # list of all subgroups of G
      N_out, # N_out(x) - function returning the minimal supergroups of x
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
	     # (see MinimalSupergroupsLattice in the Reference Manual for details
      socle, # the socle of the group G
     marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
      queue, # list used to simulate a queue data structure 
     result, # the list of essential subgroups to be returned as result of the function
x, sg, i, c; # auxiliary variables used in for loops etc.

# it is assumed that "minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return minsl[x]{[1..Size(minsl[x])]}[1];
end;

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: EssentialSubgroups(<G>) - where <G> is a finite abelian group\n");
fi;

# treat the G = TrivialSubgroup(G) extreme case
if G = TrivialSubgroup(G) then return [ G ]; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the minimal supergroups relation (before any use of N_out !!!)
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

# the socle of G is the first essential subgroup
socle := Position(sgl, Socle(G));
queue := [ socle ];
marked[socle] := true;

result := [];

# Breadth-First Search for the extensions of Socle(G)
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(result, x);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg]; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of all superfluous subgroups of G.

SuperfluousSubgroups := function(G)
local   sgl, # list of all subgroups of G
      N_out, # N_out(x) - function returning the maximal subgroups of x
      maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice 
	     # (see MaximalSubgroupsLattice in the Reference Manual for details
    radical, # the radical of the group G
     marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
      queue, # list used to simulate a queue data structure 
     result, # the list of superfluous subgroups to be returned as result of the function
x, sg, i, c; # auxiliary variables used in for loops etc.

# it is assumed that "maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return maxsl[x]{[1..Size(maxsl[x])]}[1];
end;

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: SuperfluousSubgroups(<G>) - where <G> is a finite abelian group\n");
fi;

# treat the G = TrivialSubgroup(G) extreme case
if G = TrivialSubgroup(G) then return [ G ]; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the maximal subgroups relation (before any use of N_out !!!)
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

# the radical of G is the first superfluous subgroup
radical := Position(sgl, Intersection(MaximalSubgroups(G)));
queue := [ radical ];
marked[radical] := true;

result := [];

# Breadth-First Search for the subgroups of Radical(G)
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(result, x);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg]; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of all closed subgroups of G.

ClosedSubgroups := function(G) 

local   sgl, # list of all subgroups of G
      N_out, # N_out(x) - function returning the minimal supergroups of x
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
	     # (see MinimalSupergroupsLattice in the Reference Manual for details
      socle, # the socle of the group G
     marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
      queue, # list used to simulate a queue data structure 
     result, # the list of closed subgroups to be returned as result of the function
        atm, # a list used to store (for each subgroup) the number of atoms included
x, sg, t, c, atom, minsg, g; # auxiliary variables used in for loops etc.


# it is assumed that "minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return minsl[x]{[1..Size(minsl[x])]}[1];
end;

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: ClosedSubgroups(<G>) - where <G> is a finite abelian group\n");
fi;

# treat the trivial case
if G = TrivialSubgroup(G) then  return [ G ]; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the minimal supergroups relation (before any use of N_out !!!)
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));

# the trivial subgroup
t := Position(sgl, TrivialSubgroup(G));

atm := List([1..Size(sgl)], c -> []);

# every atom contains itself
for atom in N_out(t) do atm[atom] := [ atom ]; od;

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

queue := [ t ];

while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in N_out(x) do
    UniteSet(atm[sg], atm[x]);
    if not marked[sg] then 
      Add(queue, sg);
      marked[sg] := true;
    fi;
  od;
  atm[x] := Size(atm[x]);
od;

return sgl{Filtered([1..Size(sgl)], x -> N_out(x) = [] or Minimum(atm{N_out(x)}) > atm[x])};

end;

##########################################################################################################
## Returns a list of closures of A in G.

Closures := function(G, A)

local   sgl, # list of all subgroups of G
      N_out, # N_out(x) - function returning the minimal supergroups of x
      minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
	     # (see MinimalSupergroupsLattice in the Reference Manual for details
     marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
      queue, # list used to simulate a queue data structure 
     result, # the list of closed subgroups to be returned as result of the function
        atm, # a list used to store (for each subgroup) the number of atoms included
x, sg, t, c, atom, minsg, g, a; # auxiliary variables used in for loops etc.


# it is assumed that "minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return minsl[x]{[1..Size(minsl[x])]}[1];
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: Closures(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# treat the trivial case
if G = TrivialSubgroup(G) then  return [ G ]; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the minimal supergroups relation (before any use of N_out !!!)
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));

# the trivial subgroup
t := Position(sgl, TrivialSubgroup(G));

atm := List([1..Size(sgl)], c -> []);

# every atom contains itself
for atom in N_out(t) do atm[atom] := [ atom ]; od;

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

queue := [ t ];

while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in N_out(x) do
    UniteSet(atm[sg], atm[x]);
    if not marked[sg] then 
      Add(queue, sg);
      marked[sg] := true;
    fi;
  od;
  atm[x] := Size(atm[x]);
od;

marked := List([1..Size(sgl)], c -> false);

a := Position(sgl, A);
queue := [ a ];
marked[a] := true;

result := [];

# Determine the maximal essential extensions of A
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg] and atm[x] = atm[sg]; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
  if Filtered(N_out(x), function(sg) return atm[x] = atm[sg]; end) = [] then Add(result, x); fi;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of coclosures of A in G.

Coclosures := function(G, A)

local   sgl, # list of all subgroups of G
      N_out, # N_out(x) - function returning the minimal supergroups of x
      maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice 
	     # (see MaximalSubgroupsLattice in the Reference Manual for details
     marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
      queue, # list used to simulate a queue data structure 
     result, # the list of closed subgroups to be returned as result of the function
       datm, # a list used to store (for each subgroup) the number of dual atoms in which is included
x, sg, t, c, datom, g, a; # auxiliary variables used in for loops etc.


# it is assumed that "maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return maxsl[x]{[1..Size(maxsl[x])]}[1];
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: Coclosures(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# treat the trivial case
if G = TrivialSubgroup(G) then  return [ G ]; fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the maximal subgroups relation (before any use of N_out !!!)
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

g := Position(sgl, G);

datm := List([1..Size(sgl)], c -> []);

# every dual atom is contained in itself
for datom in N_out(g) do datm[datom] := [ datom ]; od;

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

queue := [ g ];

while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in N_out(x) do
    UniteSet(datm[sg], datm[x]);
    if not marked[sg] then 
      Add(queue, sg);
      marked[sg] := true;
    fi;
  od;
  datm[x] := Size(datm[x]);
od;

marked := List([1..Size(sgl)], c -> false);

a := Position(sgl, A);
queue := [ a ];
marked[a] := true;

result := [];

# Determine the minimal coessential subgroups of A
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg] and datm[x] = datm[sg]; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
  if Filtered(N_out(x), function(sg) return datm[x] = datm[sg]; end) = [] then Add(result, x); fi;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of all coessential subgroups of A in G. 

CoessentialSubgroups := function(G, A)
local N_out, sgl, maxsl, marked, queue, sgA, S, Ca, x, sg, a;

# it is assumed that "maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return maxsl[x]{[1..Size(maxsl[x])]}[1];
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: CoessentialSubgroups(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

Ca := [ A ];

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the maximal subgroups relation (before any use of N_out !!!)
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

# we determine now all the subgroups of A
a := Position(sgl, A);
queue := [ a ];
marked[a] := true;

sgA := [];

# Breadth-First Search for the subgroups of A
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  Add(sgA, x);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg]; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
od;

S := Difference(Intersection(SuperfluousSubgroups(G), sgl{sgA}), SuperfluousSubgroups(A));

for sg in S do Ca := Union(Ca, FullSummands(A, sg)); od;

return Ca;

end;

##########################################################################################################
## Returns a list of complement subgroups of A in G.

Complements := function(G,A)
local     Za, # the set of all subgroups of G intersecting A in the trivial subgroup ("NullIntersectors")
       N_out, # N_out(x) - function returning the minimal supergroups of x       
         sgl, # list of all subgroups of G
       minsl, # list containing the minimal supergroups relation for the whole subgroup lattice 
	      # (see MinimalSupergroupsLattice in the Reference Manual for details
      marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
       queue, # list used to simulate a queue data structure 
      result, # the list of maximal "NullIntersector" subgroups to be returned as result of the function
    e, x, sg; # auxiliary variables used in for loops etc.

# it is assumed that "minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return minsl[x]{[1..Size(minsl[x])]}[1];
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: Complements(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# determine the set of all subgroups of G intersecting A in the trivial subgroup
Za := NullIntersectors(G, A);

# make Za a sorted list (a set)
Sort(Za);

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the minimal supergroups relation (before any use of N_out !!!)
minsl := MinimalSupergroupsLattice(LatticeSubgroups(G));

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

e := Position(sgl, TrivialSubgroup(G));
queue := [ e ];
marked[e] := true;

result := [];

# Determine the maximal elements of Za
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg] and sgl[sg] in Za; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
  if Filtered(N_out(x), function(sg) return sgl[sg] in Za; end) = [] then Add(result, x); fi;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of supplement subgroups of A in G.

Supplements := function(G,A)
local     Fa, # the set of all subgroups of G intersecting A in the trivial subgroup ("NullIntersectors")
       N_out, # N_out(x) - function returning the minimal supergroups of x       
         sgl, # list of all subgroups of G
       maxsl, # list containing the maximal subgroups relation for the whole subgroup lattice 
       	      # (see MaximalSubgroupsLattice in the Reference Manual for details
      marked, # boolean variable used for keeping track of already "visited" subgroups (vertices)
       queue, # list used to simulate a queue data structure 
      result, # the list of maximal "NullIntersector" subgroups to be returned as result of the function
    g, x, sg; # auxiliary variables used in for loops etc.

# it is assumed that "maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));" is called before 
# the first use of this function
N_out := function(x)
  return maxsl[x]{[1..Size(maxsl[x])]}[1];
end;

#ensure that G and A are finite abelian groups and A is a subgroup of G
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) then
  Error("Usage: Supplements(<G>, <A>) - where <G> is a finite abelian group and <A> is a subgroup of <G>\n");
fi;

# determine the set of all subgroups of G intersecting A in the trivial subgroup
Fa := FullSummands(G, A);

# make Fa a sorted list (a set)
Sort(Fa);

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# compute the maximal subgroups relation (before any use of N_out !!!)
maxsl := MaximalSubgroupsLattice(LatticeSubgroups(G));

# initialize the vector of "visited" subgroups
marked := List([1..Size(sgl)], c -> false);

g := Position(sgl, G);
queue := [ g ];
marked[g] := true;

result := [];

# Determine the minimal elements of Fa
while not IsEmpty(queue) do 
  x := Remove(queue, 1);
  for sg in Filtered(N_out(x), function(sg) return not marked[sg] and sgl[sg] in Fa; end) do 
    Add(queue, sg);
    marked[sg] := true;
  od;
  if Filtered(N_out(x), function(sg) return sgl[sg] in Fa; end) = [] then Add(result, x); fi;
od;

return sgl{result};

end;

##########################################################################################################
## Returns a list of pairs of the form [[A1, B1], [A2, B2], ..., [An, Bn]], where the subgroups
## Ai and Bi are direct summands of G (i.e. they sum up in G and intersect in 0), i = 1..n

DirectSummands := function(G) 

local n, i, sgl, l, A, B, D;

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: DirectSummands(<G>) - where <G> is a finite abelian group\n");
fi;

# obtain a list of all subgroups
sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

# n is the number of divisors of |G|
n := Size(DivisorsInt(Size(G)));

D := [];

if IsOddInt(n) then l := QuoInt(n + 1, 2);
	       else l := QuoInt(n, 2);
fi;

for i in [1..l] do
  for A in Filtered(sgl, function(g) return Size(g) = DivisorsInt(Size(G))[i]; end) do
    for B in Filtered(sgl, function(g) return Size(g) = DivisorsInt(Size(G))[n+1-i]; end) do
      if Intersection(A, B) = TrivialSubgroup(G) then
        Add(D, [A, B]);
      fi;
    od;
  od;
od;

return D;

end;

##########################################################################################################
## Returns true if B is a coessential subgroup of A in G, false otherwise. 

IsCoessential := function(G, A, B)

local epi;

#ensure that G, A and B are finite abelian groups and A is a subgroup of G; B is a subgroup of A
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) or not IsSubgroup(G, A) or not IsSubgroup(A, B) then
  Error("Usage: IsCoessential(<G>, <A>, <B>) - where <G> is a finite abelian group and B <= A <= G\n");
fi;

if G = TrivialSubgroup(G) then 
  return true;
else 
  epi := NaturalHomomorphismByNormalSubgroup(G, B);
  return IsSuperfluous(Image(epi, G), Image(epi, A));
fi;

end;

##########################################################################################################
## Returns true if G is an extending group, false otherwise.

IsExtending := function(G)

local dsum, closed;

# ensure that the group G is finite and abelian
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: IsExtending(<G>) - where <G> is a finite abelian group\n");
fi;

dsum := DirectSummands(G);

closed := ClosedSubgroups(G);

return Union(dsum{[1..Size(dsum)]}[1], dsum{[1..Size(dsum)]}[2]) = Set(closed);

end;


##########################################################################################################
## AreOrthogonal := function(G1, G2)

# returns true if G1 and G2 are perpendicular 

local sglg1, sglg2, i, sg, n1, A, B, perp;

#ensure that G1 and G2 are finite abelian groups 
if not IsGroup(G1) or not IsFinite(G1) or not IsAbelian(G1) or not IsGroup(G2) or not IsFinite(G2) or not IsAbelian(G2)then
  Error("Usage: AreOrthogonal(<G1>, <G2>) - where <G1>, <G2> are finite abelian groups\n");
fi;

perp := true;

sglg1 := List(LatticeSubgroups(G1)!.conjugacyClassesSubgroups, Representative);
sglg2 := List(LatticeSubgroups(G2)!.conjugacyClassesSubgroups, Representative);

# n1 is the number of divisors of |G1|
n1 := Size(DivisorsInt(Size(G1)));

for i in [1..n1] do
  for A in Filtered(sglg1, function(sg) return Size(sg) > 1 and Size(sg) = DivisorsInt(Size(G1))[i]; end) do
    for B in Filtered(sglg2, function(sg) return Size(sg) = Size(A); end) do
      if not IsomorphismGroups(A, B) = fail then
        perp := false;
      fi;
    od;
  od;
od;

return perp;

end;
##########################################################################################################
## AreParallel := function(G1, G2)

# returns true if G1 and G2 are parallel 

local sglg1, sglg2, sg, A, B, parallel;

#ensure that G1 and G2 are finite abelian groups 
if not IsGroup(G1) or not IsFinite(G1) or not IsAbelian(G1) or not IsGroup(G2) or not IsFinite(G2) or not IsAbelian(G2)then
  Error("Usage: AreParallel(<G1>, <G2>) - where <G1>, <G2> are finite abelian groups\n");
fi;

parallel := true;

sglg1 := List(LatticeSubgroups(G1)!.conjugacyClassesSubgroups, Representative);
sglg2 := List(LatticeSubgroups(G2)!.conjugacyClassesSubgroups, Representative);

for A in Filtered(sglg1, function(sg) return Size(sg) > 1; end) do
  if AreOrthogonal(A, G2) then parallel := false; fi;
od;
if parallel then 
  for B in Filtered(sglg2, function(sg) return Size(sg) > 1; end) do
    if AreOrthogonal(B, G1) then parallel := false; fi;
  od;
fi;

return parallel;

end;
##########################################################################################################
## TypeSubgroups := function(G)

# returns the list of type subgroups of G 

local sgl, sg, P, Y, X, typelist, okx, oky;

#ensure that G is a finite abelian group
 if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then
  Error("Usage: TypeSubgroups(<G>) - where <G> is a finite abelian group\n");
fi;

sgl := List(LatticeSubgroups(G)!.conjugacyClassesSubgroups, Representative);

typelist := [];

for P in sgl do
  oky := true;
  for Y in Filtered(sgl, function(sg) return IsSubset(sg, P) and not sg = P; end) do
    okx := false;
    for X in Filtered(sgl, function(sg) return IsSubset(Y, sg) and not sg = TrivialSubgroup(G); end) do    
      if AreOrthogonal(P, X) then okx := true; fi;
    od;
    if not okx then oky := false; fi;
  od;
  if oky and not P in typelist then Add(typelist, P); fi;
od;

return typelist;

end;
##########################################################################################################
## IsTypeSubgroup := function(G,P)

# returns true if P is a type subgroup of G 

local type;

#ensure that G is a finite abelian group 
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then
  Error("Usage: TypeSubgroups(<G>) - where <G> is a finite abelian group\n");
fi;

if P in TypeSubgroups(G) then type := true; else type := false; fi;

return type;

end;


##########################################################################################################
## IsTS := function(G) 

local dsum, ts;

# ensure that the group G is finite and abelian 
if not IsGroup(G) or not IsFinite(G) or not IsAbelian(G) then 
  Error("Usage: IsTS(<G>) - where <G> is a finite abelian group\n");
fi;

dsum := DirectSummands(G);

ts := TypeSubgroups(G);

return IsSubset(Union(dsum{[1..Size(dsum)]}[1], dsum{[1..Size(dsum)]}[2]),Set(ts));

end;
